local MINOR_VERSION = tonumber(("$Revision: 67730 $"):match("%d+"))
if MINOR_VERSION > Omen.MINOR_VERSION then Omen.MINOR_VERSION = MINOR_VERSION end

local bars
local SingleTarget = Omen:NewModule("SingleTarget", Omen.ModuleBase, "AceEvent-3.0", "AceTimer-3.0")
local Threat = LibStub("Threat-2.0")
local GUID = LibStub("LibGUIDRegistry-0.1")
local timers = {}
local table_sort = _G.table.sort
local math_abs = _G.math.abs
local OL = LibStub("AceLocale-3.0"):GetLocale("Omen")
local L = LibStub("AceLocale-3.0"):GetLocale("Omen_SingleTarget")
local GetTime = _G.GetTime
local normalizeTo100 = false
local locale = GetLocale()
local UnitExists = _G.UnitExists
local UnitIsPlayer = _G.UnitIsPlayer
local UnitPlayerControlled = _G.UnitPlayerControlled
local titleBar
local math_floor = _G.math.floor

local configOptions = {
	TPS = {
		UpdateFrequency = 0.1
	},
	Warnings = {
		Enable = true,
		EnableSound = true,
		EnableFlash = true,
		EnableShake = false,
		EnableMessage = true,
		Threshold = 90,
		SoundFile = "Fel Nova"
	},
	Columns = {
		Name = true,
		Threat = true,
		TPS = true,
		Percent = true
	},
	AggroBar = {
		Enable = true,
		Color = {r = 1, g = 0, b = 0, a = 1}
	},
	MyBar = {	
		UseCustomColor = true,
		Color = {r = 1, g = 0, b = 0, a = 1}
	},
	ShowTitleBar = true,
	ShowSelf = true
}
local Media = LibStub("LibSharedMedia-3.0")
local sounds = Media:List("sound")

local function GetLSMIndex(t, value)
	for k, v in pairs(Media:List(t)) do
		if v == value then
			return k
		end
	end
	return nil
end

local options = {
	type = "group",
	name = L["Single Target Mode"],
	desc = L["Single Target Mode"],
	args = {
		test = {
			type = "execute",
			name = OL["Show test bars"],
			desc = OL["Show test bars"],
			func = function() Omen:EnableModule("SingleTarget"); SingleTarget:Test() end
		},	
		warnings = {
			type = "group",
			name = L["Warnings"],
			desc = L["Warnings"],
			args = {
				enable = {
					type = "toggle",
					name = L["Enable Warnings"],
					desc = L["Enable Warnings"],
					order = 1,
					get = function(info)
						return SingleTarget:GetOption("Warnings.Enable")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.Enable", v)
					end
				},
				sound = {
					type = "toggle",
					name = L["Enable Sound"],
					desc = L["Enable Sound"],
					get = function(info)
						return SingleTarget:GetOption("Warnings.EnableSound")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.EnableSound", v)
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},
				flash = {
					type = "toggle",
					name = L["Enable Screen Flash"],
					desc = L["Enable Screen Flash"],
					get = function(info)
						return SingleTarget:GetOption("Warnings.EnableFlash")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.EnableFlash", v)
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},
				shake = {
					type = "toggle",
					name = L["Enable Screen Shake"],
					desc = L["Enable Screen Shake"],
					get = function(info)
						return SingleTarget:GetOption("Warnings.EnableShake")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.EnableShake", v)
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},				
				message = {
					type = "toggle",
					name = L["Enable Warning Message"],
					desc = L["Print a message to screen when you accumulate too much threat"],
					get = function(info)
						return SingleTarget:GetOption("Warnings.EnableMessage")
					end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.EnableMessage", v)
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end				
				},
				soundFile = {
					type = "select",
					name = L["Sound to play"],
					desc = L["Sound to play"],
					values = sounds,
					get = function(info) return GetLSMIndex("sound", SingleTarget:GetOption("Warnings.SoundFile")) end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.SoundFile", sounds[v])
						PlaySoundFile(Media:Fetch("sound", sounds[v]))
					end,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},
				threshold = {
					type = "range",
					name = L["Warning Threshold"],
					desc = L["Warning Threshold"],
					get = function(info) return SingleTarget:GetOption("Warnings.Threshold") end,
					set = function(info, v)
						SingleTarget:SetOption("Warnings.Threshold", v)
						SingleTarget.WarningThreshold = v
					end,
					step = 1,
					bigStep = 5,
					min = 60,
					max = 130,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				},
				test = {
					type = "execute",
					name = L["Test warnings"],
					desc = L["Test warnings"],
					func = function() SingleTarget:TestWarnings() end,
					order = 101,
					disabled = function() return not SingleTarget:GetOption("Warnings.Enable") end
				}
			}
		},
		showself = {
			type = "toggle",
			name = L["Always show self"],
			desc = L["Always show yourself on the meter"],
			get = function(info) return SingleTarget:GetOption("ShowSelf") end,
			set = function(info, v)
				SingleTarget:SetOption("ShowSelf", v)
				SingleTarget.showSelf = v
			end
		},
		showtitle = {
			type = "toggle",
			name = L["Show title bar"],
			desc = L["Show column headers on a title bar"],
			get = function(info) return SingleTarget:GetOption("ShowTitleBar") end,
			set = function(info, v)
				SingleTarget:SetOption("ShowTitleBar", v)
				if titleBar then
					titleBar = SingleTarget:ReleaseBar(titleBar)
				end
				SingleTarget.showTitle = v
				SingleTarget:ArrangeBars()
			end
		},		
		threatMode = {
			type = "toggle",
			name = L["Show tank as 100%"],
			desc = L["Show tank as 100%, rather than 0%"],
			get = function(info) return SingleTarget:GetOption("NormalizeTo100") end,
			set = function(info, v)
				SingleTarget:SetOption("NormalizeTo100", v)
				normalizeTo100 = v
			end
		},
		columns = {
			type = "group",
			name = L["Columns"],
			desc = L["Select which columns to show"],
			args = {
				name = {
					type = "toggle",
					name = L["Name"],
					desc = L["Show the Name column"],
					get = function(info) return SingleTarget:GetOption("Columns.Name") end,
					set = function(info, v)
						SingleTarget:SetOption("Columns.Name", v)
						SingleTarget:UpdateBarLabels()
					end
				},
				tps = {
					type = "toggle",
					name = L["TPS"],
					desc = L["Show the TPS column"],
					get = function(info) return SingleTarget:GetOption("Columns.TPS") end,
					set = function(info, v)
						SingleTarget:SetOption("Columns.TPS", v)
						SingleTarget:UpdateBarLabels()
					end
				},
				tpqfreq = {
					type = "range",
					name = L["TPS Update Frequency"],
					desc = L["TPS Update Frequency"],
					min = 0.05,
					max = 5.0,
					step = 0.05,
					bigStep = 0.05,
					get = function(info) return SingleTarget:GetOption("TPS.UpdateFrequency") end,
					set = function(info, v)
						SingleTarget:SetOption("TPS.UpdateFrequency", v)
						SingleTarget:ReInitTPSTimer()
					end
				},
				threat = {
					type = "toggle",
					name = L["Threat"],
					desc = L["Show the Threat column"],
					get = function(info) return SingleTarget:GetOption("Columns.Threat") end,
					set = function(info, v)
						SingleTarget:SetOption("Columns.Threat", v)
						SingleTarget:UpdateBarLabels()
					end
				},
				pct = {
					type = "toggle",
					name = L["Percent"],
					desc = L["Show the Percent column"],
					get = function(info) return SingleTarget:GetOption("Columns.Percent") end,
					set = function(info, v)
						SingleTarget:SetOption("Columns.Percent", v)
						SingleTarget:UpdateBarLabels()
					end
				}
			}
		},
		yourbar = {
			type = "group",
			name = L["Your Bar"],
			desc = L["Your Bar"],
			args = {
				color = {
					type = "color",
					name = OL["Color"],
					desc = L["The color of your bar's background"],
					hasAlpha = true,
					get = function(info)
						local c = SingleTarget:GetOption("MyBar.Color")
						return c.r, c.g, c.b, c.a
					end,
					set = function(info, r, g, b, a)
						SingleTarget:SetOption("MyBar.Color.r", r)
						SingleTarget:SetOption("MyBar.Color.g", g)
						SingleTarget:SetOption("MyBar.Color.b", b)
						SingleTarget:SetOption("MyBar.Color.a", a)
						local pguid = UnitGUID("player")
						for _, bar in pairs(bars) do
							if bar.guid == pguid then
								bar:SetColor(r, g, b, a)
								break
							end
						end
					end,
					disabled = function() return not SingleTarget:GetOption("MyBar.UseCustomColor") end
				},
				customcolor = {
					type = "toggle",
					name = L["Use custom color"],
					desc = L["Use a custom color for your bar in the display list"],
					get = function() return SingleTarget:GetOption("MyBar.UseCustomColor") end,
					set = function(info, v)
						SingleTarget:SetOption("MyBar.UseCustomColor", v)
						local pguid = UnitGUID("player")
						for _, bar in pairs(bars) do
							if bar.guid == pguid then
								if v then
									bar:SetColor(
										SingleTarget:GetOption("MyBar.Color.r"),
										SingleTarget:GetOption("MyBar.Color.g"),
										SingleTarget:GetOption("MyBar.Color.b"),
										SingleTarget:GetOption("MyBar.Color.a")									
									)
								else
									bar:SetClass()
								end
								break
							end
						end
					end
				}
			}
		},
		aggrobar = {
			type = "group",
			name = L["Aggro Bars"],
			desc = L["Aggro Bars"],
			args = {
				show = {
					type = "toggle",
					name = L["Show bar"],
					desc = L["Show bar"],
					get = function(info) return SingleTarget:GetOption("AggroBar.Enable") end,
					set = function(info, v)
						SingleTarget:SetOption("AggroBar.Enable", v)
						SingleTarget.showAggroBar = v
						SingleTarget:ArrangeBars()
					end
				},
				color = {
					type = "color",
					name = L["Color"],
					desc = L["The color of the aggro gain bar"],
					get = function(info)
						local c = SingleTarget:GetOption("AggroBar.Color")
						return c.r, c.g, c.b, c.a
					end,
					set = function(info, r, g, b, a)
						SingleTarget:SetOption("AggroBar.Color.r", r)
						SingleTarget:SetOption("AggroBar.Color.g", g)
						SingleTarget:SetOption("AggroBar.Color.b", b)
						SingleTarget:SetOption("AggroBar.Color.a", a)
						local bar = SingleTarget:AcquireBar("AGGRO", true)
						if bar then
							bar:SetColor(r,g,b, a)
						end
					end				
				}
			}
		},
		
	}
}

local function sortBars(a, b)
	if a.isTitle then return true end
	if b.isTitle then return false end
	return (a:GetValue() > b:GetValue())
end
local selfThreat, aggroThreat, aggroGUID = 0, nil, nil
local hostileGUID, hostileUnit, playerGUID = nil, "target", nil

function SingleTarget:OnInitialize()
	self:Super("OnInitialize")
	self.icon = select(3, GetSpellInfo(53))
	self:RegisterConfigDefaults(configOptions)
	self:RegisterOptions(options)
	self.threatUpdatedAt = 0
	self.WarningThreshold = self:GetOption("Warnings.Threshold")
	self.showSelf = self:GetOption("ShowSelf")
	self.showAggroBar = self:GetOption("AggroBar.Enable")
	self.showTitle = self:GetOption("ShowTitleBar")
	locale = GetLocale()
end

function SingleTarget:OnEnable()
	self:Super("OnEnable")
	bars = self.bars
	
	Threat.RegisterCallback(self, "ThreatUpdated")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
	self:RegisterEvent("UNIT_TARGET")
	-- self:RegisterEvent("UNIT_FACTION", "PLAYER_TARGET_CHANGED")
	playerGUID = UnitGUID("player")
	
	self:PLAYER_TARGET_CHANGED()
	if timers.OmenCancelTPSUpdate then
		self:CancelTimer(timers.OmenCancelTPSUpdate, true)
	end
	timers.OmenCancelTPSUpdate = nil
	normalizeTo100 = self:GetOption("NormalizeTo100")
end

local barOptions = {}
function SingleTarget:GetBarOptions()
	bars = self.bars
	local options = barOptions
	for i = 1, #options do
		tremove(options)
	end
	local used = 0
	if self:GetOption("Columns.Name") then
		tinsert(options, "Name")
		tinsert(options, "LEFT")
		tinsert(options, 40)
		used = used + 40
	end
	if self:GetOption("Columns.TPS") then
		tinsert(options, "TPS")
		tinsert(options, "LEFT")
		tinsert(options, 20)
		used = used + 20
	end
	if self:GetOption("Columns.Threat") then
		tinsert(options, "Threat")
		tinsert(options, "RIGHT")
		tinsert(options, 20)
		used = used + 20
	end
	if self:GetOption("Columns.Percent") then
		tinsert(options, "%")
		tinsert(options, "RIGHT")
		tinsert(options, 20)
		used = used + 20
	end
	for i = 3, #options, 3 do
		options[i] = options[i] * (100 / used)
	end
	return options
end

function SingleTarget:UpdateBarLabels()
	local options = self:GetBarOptions()
	for k, v in pairs(bars) do
		v:ClearLabels()
		v:SetLabels(unpack(options))
	end
end

function SingleTarget:OnDisable()
	self:Super("OnDisable")
	if titleBar then
		titleBar = nil
	end
	Threat.UnregisterCallback(self, "ThreatUpdated")
end

function SingleTarget:Hint()
	return L["Single Target Mode\n|cffffffffShows raid threat on your target|r"]
end

function SingleTarget:UpdateLayout()
	self:ArrangeBars()
end

local lastUpdatedAt = 0
function SingleTarget:UpdateBar(guid, threat)
	local bar, isNew = self:AcquireBar(guid)
	if isNew then
		bar:SetLabels(unpack(self:GetBarOptions()))
		bar:SetClass()
		if bar.guid == UnitGUID("player") and self:GetOption("MyBar.UseCustomColor") then
			bar:SetColor(
				self:GetOption("MyBar.Color.r"),
				self:GetOption("MyBar.Color.g"),
				self:GetOption("MyBar.Color.b"),
				self:GetOption("MyBar.Color.a")
			)
		end
		bar:SetLabel("Name", Threat.GUIDNameLookup[guid])
	end
	bar:SetLabel("Threat", bar:ConvertToK(threat))
	bar:SetValue(threat)
	if GetTime() - lastUpdatedAt > 0.25 then
		aggroThreat = aggroThreat or Threat:GetMaxThreatOnTarget(guid)
		lastUpdatedAt = GetTime()
		self:ArrangeBars()
	end
end

local function getAggroBar(self)
	local bar, isNew = self:AcquireBar("AGGRO")
	if isNew then
		bar:SetLabels(unpack(self:GetBarOptions()))
		bar:SetColor(
			self:GetOption("AggroBar.Color.r"),
			self:GetOption("AggroBar.Color.g"),
			self:GetOption("AggroBar.Color.b"),
			self:GetOption("AggroBar.Color.a")
		)
		bar:SetLabel("Name", ("> %s <"):format(L["Pull Aggro"]))
		bar:SetLabel("TPS", "")
	end
	return bar
end

function SingleTarget:ArrangeBars()
	if self.showTitle then
		local newTitle
		titleBar, newTitle = self:AcquireBar("TITLE")
		if newTitle then
			titleBar.isTitle = true
			titleBar:SetValue(1)
			titleBar:SetColor(0,0,0,0)
			titleBar:SetLabels(unpack(self:GetBarOptions()))
			titleBar:SetLabel("Name", L["Name"])
			titleBar:SetLabel("TPS", L["TPS"])
			titleBar:SetLabel("Threat", L["Threat"])
			titleBar:SetLabel("%", L["Percent"])
		end
	end
	local showSelf, selfShown = self.showSelf, false
	local bar = getAggroBar(self)
	if bar and #bars > 0 and self.showAggroBar and hostileUnit then
		if Threat:UnitInMeleeRange(hostileUnit) then
			bar:SetValue(aggroThreat * 1.1)
			bar:SetLabel("Threat", bar:ConvertToK(aggroThreat * 1.1))
		else
			bar:SetValue(aggroThreat * 1.3)
			bar:SetLabel("Threat", bar:ConvertToK(aggroThreat * 1.3))
		end
	elseif bar then
		bar:SetValue(0)
	end
	
	self:ResetBars()
	table_sort(bars, sortBars)
	local maxThreat = nil
	for i = 1, 80 do		-- A raid of 40 hunters!
		local bar = bars[i]
		if not bar then break end
		
		local hide = (bar.isPet and not Omen:GetOption("Classes.PET")) or (bar.class and not Omen:GetOption("Classes." .. bar.class))
		
		local spaceLeft, barsLeft = self:BarSpaceLeft()
		
		if bar.guid == playerGUID then
			selfShown = true
			hide = false
		elseif showSelf and not selfShown then
			barsLeft = barsLeft - 1
		end
		if bar.isTitle then
			if self.showTitle then
				self:AddBar(bar)
			end
		elseif bar:GetValue() > 0 and not hide and barsLeft > 0 and self:AddBar(bar) then
			maxThreat = maxThreat or bar:GetValue()
			local pct = bar:GetValue() / maxThreat
			if aggroThreat == 0 then
				aggroThreat = bar:GetValue()
			end
			local visPct = bar:GetValue() / aggroThreat
			bar:SetPercent(pct)
			if normalizeTo100 then
				bar:SetLabel("%", math_floor(visPct * 100) .. "%")
			else
				bar:SetLabel("%", math_floor((visPct - 1) * 100) .. "%")
				-- bar:SetLabel("%", ("%+2.0f%%"):format((visPct-1)*100))
			end
		else
			bar.frame:Hide()
		end
	end
end

function SingleTarget:ThreatUpdated(event, srcGUID, dstGUID, threat)
	if not hostileUnit then return end
	if self.testing then
		self:ReleaseBars()
		self.testing = false
	end
	
	if (hostileGUID and dstGUID == hostileGUID) then
		aggroThreat = aggroGUID and ((srcGUID == aggroGUID) and threat or Threat:GetThreat(aggroGUID, dstGUID)) or Threat:GetMaxThreatOnTarget(dstGUID) or 1
		selfThreat = (srcGUID == playerGUID) and threat or Threat:GetThreat(playerGUID, dstGUID) or 0
		
		self:CheckWarn(srcGUID, dstGUID, threat)
		self:UpdateBar(srcGUID, threat)
	end
	
	self.threatUpdatedAt = GetTime()
	if not timers.OmenCancelTPSUpdate then
		timers.OmenCancelTPSUpdate = self:ScheduleRepeatingTimer("CancelTPSUpdate", 6)
		timers.OmenTPSUpdate = self:ScheduleRepeatingTimer("UpdateTPS", self:GetOption("TPS.UpdateFrequency"))
	end
end

function SingleTarget:ReInitTPSTimer()
	if timers.OmenTPSUpdate then
		self:CancelTimer(timers.OmenTPSUpdate)
		timers.OmenTPSUpdate = self:ScheduleRepeatingTimer("UpdateTPS", self:GetOption("TPS.UpdateFrequency"))
	end
end

function SingleTarget:CancelTPSUpdate()
	if GetTime() - self.threatUpdatedAt > 5 then
		if timers.OmenTPSUpdate then
			self:CancelTimer(timers.OmenTPSUpdate, true)
			timers.OmenTPSUpdate = nil
		end
		self:CancelTimer(timers.OmenCancelTPSUpdate)
		timers.OmenCancelTPSUpdate = nil
	end
end

function SingleTarget:UpdateTPS()
	if not hostileUnit then return end
	local target_guid = UnitGUID(hostileUnit)
	if not target_guid then return end
	for k, bar in pairs(bars) do
		if not bar.isTitle and bar.frame:IsVisible() and bar.guid ~= "AGGRO" then
			bar:SetLabel("TPS", ("%2.0f"):format(Threat:GetTPS(bar.guid, target_guid)))
		end
	end
end

function SingleTarget:DisplayThreatForGUID(guid)
	for k, v in Threat:IterateGroupThreatForTarget(guid) do
		self:UpdateBar(k, v)
	end
end

local function isValidThreatTarget(unit)
	if not UnitExists(unit) or UnitIsPlayer(unit) or UnitPlayerControlled(unit) then
		return false 
	end
	return true
end

function SingleTarget:TargetTargetCheck()
	if not timers.TargetSwitch and UnitGUID(self.aggroUnit) ~= aggroGUID then
		self:UNIT_TARGET(nil, "target")
	end
end

function SingleTarget:PLAYER_TARGET_CHANGED()
	hostileGUID = nil
	hostileUnit = nil
	aggroGUID = nil
	aggroThreat = nil
	
	if timers.TargetTargetCheck then
		self:CancelTimer(timers.TargetTargetCheck)
		timers.TargetTargetCheck = nil
	end
	
	if isValidThreatTarget("target") then
		hostileUnit = "target"
		self.aggroUnit = "targettarget"
	elseif UnitExists("targettarget") and isValidThreatTarget("targettarget") then
		hostileUnit = "targettarget"
		self.aggroUnit = "targettargettarget"
		timers.TargetTargetCheck = self:ScheduleRepeatingTimer("TargetTargetCheck", 0.5)
	end
	
	self:ClearBars()
	if not hostileUnit then
		self:SetTitle(nil)
		return
	end
	hostileGUID = UnitGUID(hostileUnit) 
	aggroGUID = UnitGUID(self.aggroUnit)
	
	aggroThreat = Threat:GetThreat(aggroGUID, hostileGUID) or 1
	selfThreat = Threat:GetThreat(UnitGUID("player"), hostileGUID) or 0
	self:DisplayThreatForGUID(hostileGUID)
	self:UpdateTPS()
	self:SetTitle(UnitName(hostileUnit))
	
	self:UNIT_TARGET(nil, "target")
end

function SingleTarget:CheckWarn(s, d, t)
	if s ~= playerGUID and s ~= aggroGUID then return end
	-- if self.playerGUID = self.aggroGUID then return end
	
	if not self:GetOption("Warnings.Enable") then return end
	
	local pct = selfThreat / aggroThreat * 100
	if pct > self.WarningThreshold and not self.warned and aggroGUID and playerGUID ~= aggroGUID and aggroThreat > 5 then
		local msg = nil
		if self:GetOption("Warnings.EnableMessage") then
			if (locale == "zhCN") then
				msg = (L["Passed %s%% of %s's threat!"]):format(Threat.GUIDNameLookup[aggroGUID], self.WarningThreshold)
			else
				msg = (L["Passed %s%% of %s's threat!"]):format(self.WarningThreshold, Threat.GUIDNameLookup[aggroGUID])
			end
		end
		Omen:Warn(
			self:GetOption("Warnings.EnableSound") and Media:Fetch("sound", self:GetOption("Warnings.SoundFile")),
			self:GetOption("Warnings.EnableFlash"),
			self:GetOption("Warnings.EnableShake"),
			msg
		)
		self.warned = true
	elseif pct < (self.WarningThreshold - 5) and self.warned then
		self.warned = false
	end
end

function SingleTarget:TestWarnings()
	Omen:Warn(
		SingleTarget:GetOption("Warnings.EnableSound") and Media:Fetch("sound", SingleTarget:GetOption("Warnings.SoundFile")),
		SingleTarget:GetOption("Warnings.EnableFlash"),
		SingleTarget:GetOption("Warnings.EnableShake"),
		SingleTarget:GetOption("Warnings.EnableMessage") and "Warning Test" or nil
	)
end

function SingleTarget:SwitchAggroTarget()
	local n = UnitName(self.aggroUnit)
	if n then
		aggroGUID = UnitGUID(self.aggroUnit)
	end
	timers.TargetSwitch = nil
end

function SingleTarget:UNIT_TARGET(event, unit)
	if unit ~= "target" then return end
	if (not hostileUnit and isValidThreatTarget("targettarget")) or (hostileUnit == "targettarget" and hostileGUID ~= UnitGUID(hostileUnit)) then
		self:PLAYER_TARGET_CHANGED()
		return
	end
	if not hostileUnit then return end
	
	if timers.TargetSwitch then
		self:CancelTimer(timers.TargetSwitch, true)
		timers.TargetSwitch = nil
	end
	if not UnitCastingInfo(hostileUnit) or not aggroGUID then
		timers.TargetSwitch = self:ScheduleTimer("SwitchAggroTarget", 0.5)
	end
end

function SingleTarget:Test()
	self:ReleaseBars()
	
	aggroThreat = 0 
	self.testing = true
	for i = 1, 25 do
		local threat = random() * 10000
		aggroThreat = max(aggroThreat, threat)
		self:UpdateBar(i, threat)
	end
	self:ArrangeBars()
end
